#!/usr/bin/env python3

""" A more long-term testing of pingcheck

Runs pingcheck parsing its log output; goes online and offline repeatedly to see if
  pinging stays correct

see also: connd/stresstest_pingcheck.py

Christian Herdtweck, Intra2net, January 2015
(c) Intra2net AG 2015
"""

# Version History
# 15/01/15 Christian Herdtweck: started creation
# 21/1/15 ch: running 3rd long-term test, seems to work nicely now

import subprocess
import sys
import select
from re import match as regexp
from time import sleep
from datetime import timedelta, datetime as dt
from warnings import warn
from random import randint
import logging
from os.path import expanduser, dirname, isdir
from os import mkdir, EX_OK
from connd_state import ConndState
#import signal

logger_base = logging.getLogger('long_term_test')

########################################################################
### CONSTANTS
########################################################################

version = '1.1'
#default_pingcheck_bin  = '/usr/intranator/bin/pingcheck'
default_pingcheck_bin  = '/root/git/pingcheck/build/src/pingcheck'
default_tell_connd_bin = '/usr/intranator/bin/tell-connd'
default_output_prefix = './long_term_test/{0:%d%m%y_%H%M}'

# for parsing log messages
LEVEL_DEBUG  = 'dbug'
LEVEL_INFO   = 'info'
LEVEL_WARN   = 'warn'
LEVEL_NOTICE = 'note'
LEVEL_ERROR  = 'err '

text_to_level = {}
text_to_level['INFO'] = LEVEL_INFO
text_to_level['DBUG'] = LEVEL_DEBUG
text_to_level['WARN'] = LEVEL_WARN
text_to_level['NOTE'] = LEVEL_NOTICE
text_to_level['ERR']  = LEVEL_ERROR

# duration in which to remain online/offline
MIN_DURATION_SAME_STATE = 10   # [seconds]
MAX_DURATION_SAME_STATE = 50   # [seconds]

########################################################################
#### for config file
########################################################################

PING_INTERVAL = 3    # [seconds]
config_file_pattern = 'long_term_test_{0:%d%m%y_%H%M%S}.conf'
conf_start = """# config file for pingcheck testing

default-source-network-interface=eth1
nameserver=127.0.0.1
hosts-down-limit={hosts_down_limit:d}
ping-fail-limit=25
status-notifier-cmd=echo host status is ${status}
link-up-interval=2
ping-reply-timeout=3
"""
conf_host = """
[host]
name={0}
interval={1:d}
ping-protocol=ICMPv4
port=80
source-network-interface=default
"""
host_list = (\
        'intranator.m.i2n', \
        'intraroute.m.i2n', \
        'websync.m.i2n', \
        'aqua.m.i2n', \
        'intratdsl.m.i2n', \
        'beholder.m.i2n', \
        'eagle.m.i2n', \
        'osdist.m.i2n', \
        'tokyo.m.i2n', \
        'blizzard.m.i2n', \
        'phonehelp.m.i2n', \
        'virt-dmz.m.i2n', \
        'virt01.m.i2n', \
        'virt02.m.i2n', \
        'tornado.n.i2n', \
        'auth.m.i2n', \
        'forum.intra2net.com', \
        'update.intranator.com', \
        'demo-intranator.intra2net.com', \
        'intra2net.com', \
        'partner.intra2net.com', \
        'demo-securitygateway.intra2net.com', \
        'demo-networksecurity.intra2net.com', \
    )

def create_conf_file(file_prefix):
    """ create config file using constants above """

    file_name = file_prefix + '_config.conf'

    n_hosts = len(host_list)

    with open(file_name, 'wt') as file_handle:
        file_handle.write(conf_start.format(hosts_down_limit=int(n_hosts * 0.7), status='{status}'))
        for host in host_list:
            file_handle.write(conf_host.format(host, PING_INTERVAL))

    return file_name, n_hosts
#end: create_conf_file


########################################################################
### Exceptions
########################################################################

def flags_to_str(flags):
    text_parts = []
    if flags & select.POLLIN:
        text_parts.append('IN')
    if flags & select.POLLOUT:
        text_parts.append('OUT')
    if flags & select.POLLPRI:
        text_parts.append('PRI')
    if flags & select.POLLERR:
        text_parts.append('ERR')
    if flags & select.POLLHUP:
        text_parts.append('HUP')
    if flags & select.POLLNVAL:
        text_parts.append('INVALID')
    return ','.join(text_parts)

class PollFlagsException(Exception):
    def __init__(self, flags, proc_name):
        super(PollFlagsException, self).__init__('Unexpected flags from polling {0}: {1}'.format(\
                proc_name, flags_to_str(flags)))

class PollHupException(Exception):
    def __init__(self, proc_name):
        super(PollHupException, self).__init__(\
                'Received HUP from polling, {0} was probably killed!'.format(proc_name))

class TooManyWarnings(Exception):
    def __init__(self):
        super(TooManyWarnings, self).__init__('Too many warnings / failures!')

########################################################################
### PARSER
########################################################################

class ProcOutputParser:
    """ parser for output from pingcheck process via pipes

    uses select and polling
    should never silently swallow log output, main function parse should yield all lines
      or emit errors
    """

    _out = None
    _err = None
    _poller = None
    _flags = select.POLLIN | select.POLLPRI | select.POLLHUP # | select.POLLERR
    _regexp = '(.+)\s+(INFO|ERR|WARN|DBUG|NOTE)\s+pingcheck\[\d+\]: (.+)\s*'
    _out_fileno = None
    _err_fileno = None

    SRC_OUT = 'stdout'
    SRC_ERR = 'stderr'
    SRC_UNK = 'unknown'

    def __init__(self, pipe_out, pipe_err):
        self._out = pipe_out
        self._err = pipe_err
        self._poller = select.poll()
        self._poller.register(pipe_out, self._flags)
        self._poller.register(pipe_err, self._flags)
        self._out_fileno = pipe_out.fileno()
        self._err_fileno = pipe_err.fileno()

    def _read(self):
        while True:
            result = self._poller.poll(10)
            for file_no, flags in result:
                if not (flags & self._flags):
                    raise PollFlagsException(flags, 'pingcheck')

                if flags & select.POLLHUP:
                    raise PollHupException('pingcheck')

                if file_no == self._out_fileno:
                    src = self.SRC_OUT
                    line = self._out.readline()
                elif file_no == self._err_fileno:
                    src = self.SRC_ERR
                    line = self._err.readline()
                else:
                    src = self.SRC_UNK
                    line = ''

                if not line:
                    continue
                # if reach here, we have a new line of text

                yield src, line.strip()
    #end: ProcOutputParser._read

    def parse(self):
        for src, line in self._read():
            try:
                timestamp, level, message = regexp(self._regexp, line).groups()
                yield src, (timestamp, text_to_level[level], message)
            except:
                yield src, line

    def __iter__(self):
        return self.parse()

#end: class ProcOutputParser

########################################################################
### TESTER
########################################################################

class Tester:
    """ """

    _line_iter = None
    _last_request = None
    _last_reply = None
    _n_requests = None
    _n_replies = None
    _next_state_change = None
    _tell_connd_bin = None
    _prev_state_changes = None
    _fails = None
    _log = None
    _log_ping = None
    max_fails = None

    def __init__(self, line_iter, tell_connd_bin, max_fails):
        self._line_iter = line_iter
        self._last_request = {}
        self._last_reply = {}
        self._n_requests = {}
        self._n_replies = {}
        self._tell_connd_bin = tell_connd_bin
        self._prev_state_changes = []
        self._fails = []
        self.max_fails = max_fails

        self._log = logger_base.getChild('Tester')
        self._log_ping = {}
        self._log_ping[ProcOutputParser.SRC_OUT] = logger_base.getChild('pingOut')
        self._log_ping[ProcOutputParser.SRC_ERR] = logger_base.getChild('pingErr')


    def test(self):
        # preparations
        current_state = self._is_online()
        while current_state is None:
            sleep(1)
            current_state = self._is_online()
        self._log.info('Start with state online = {0}'.format(current_state))
        self._prev_state_changes.append((current_state, dt.now()))
        self._schedule_next_state_change()

        # start looping over output from observed process
        for src, text_info in self._line_iter:
            if isinstance(text_info, str):
                # ProcOutputParser could not parse
                msg = 'unparsed: {0}'.format(text_info)
                if text_info.startswith("RECV BUF ("):
                    level = LEVEL_DEBUG
                else:
                    level = LEVEL_WARN
            else:
                text_time, level, message = text_info
                # first of all: write to text log so nothing is lost
                #msg = '{0} {1} (orig time {2})'.format(level, message, text_time)
                msg = '{0} {1}'.format(level, message)

            if level in (LEVEL_WARN, LEVEL_ERROR, LEVEL_NOTICE):
                # most common warning/error is only logged at debug level
                if message.endswith('could not be resolved.'):
                    self._log_ping[src].debug(msg)
                    continue
                elif message.endswith('fail sending ping data. Network is unreachable'):
                    self._log_ping[src].debug(msg)
                    continue
                elif message.endswith('fail sending ping data. Operation not permitted'):
                    self._log_ping[src].debug(msg)
                    continue
                else:   # unexpected errors and warnings from pingcheck are logged at info level
                    self._log_ping[src].info(msg)
                    continue
            else:   # normal messages are logged at debug level
                self._log_ping[src].debug(msg)

            # now deal with message contents
            if self._handle_request(message):
                continue
            if self._handle_reply(message):
                continue
            if self._handle_change_state():
                continue
            if self._handle_new_ip(message):
                continue
        #end: for all lines from pingcheck output
    #end: Tester.test


    def _handle_new_ip(self, message):
        matches = regexp('- (\d+\.\d+\.\d+\.\d+) \[\d+s\]', message)
        if matches is None:
            matches = regexp('Host is already an IP: (\d+\.\d+\.\d+\.\d+)', message)
        if matches:
            ip = matches.groups()[0]
            self._log.info('remember new IP {0}'.format(ip))
            self._last_request[ip] = dt.now()
            self._last_reply[ip] = dt.now()
            self._n_requests[ip] = 0
            self._n_replies[ip] = 0
            return True
        return False


    def _handle_request(self, message):
        matches = regexp('(\d+\.\d+\.\d+\.\d+): sending ping', message)
        if matches is not None:
            ip = matches.groups()[0]
            self._log.info('pinging IP {0}'.format(ip))
            self._last_request[ip] = dt.now()
            self._n_requests[ip] += 1
            return True
        return False


    def _handle_reply(self, message):
        matches = regexp('(\d+\.\d+\.\d+\.\d+): Received reply', message)
        if matches is not None:
            ip = matches.groups()[0]
            self._log.info('received reply from {0}'.format(ip))
            self._last_reply[ip] = dt.now()
            self._n_replies[ip] += 1
            return True
        return False


    def _handle_change_state(self):
        if dt.now() > self._next_state_change:
            is_online = self._is_online()
            if is_online is None:
                return False   # try again later
            elif is_online:
                self._ensure_all_pinged()
                new_val = False
            else:
                new_val = True
            if self._set_online(new_val):
                self._schedule_next_state_change()
                return True
            # if call to tell_connd fails, try again later
        return False
    #end: _handle_change_state


    def _schedule_next_state_change(self):
        n_secs = randint(MIN_DURATION_SAME_STATE, MAX_DURATION_SAME_STATE)
        self._next_state_change = dt.now() + timedelta(seconds=n_secs)
        self._log.info('schedule next state change for {0:%H:%M:%S} (in {1:d}s)'.format(\
                self._next_state_change, n_secs))
    #end: Tester._schedule_next_state_change


    def _is_online(self):
        """ determine state using ConndState """
        state = ConndState.get_state(self._tell_connd_bin)
        if not isinstance(state, ConndState):
            err_code, output_lines = state
            self._warn('Non-zero exit status asking tell-connd for its --status: {0}'.format(\
                        err_code, output_lines[0]))
            return None

        return len(state.connections) > 0
    #end: Tester._is_online


    def _set_online(self, new_val):
        """ run tell-connd --online/offline P1

        returns true if afterwards are in requested state
        """
        if new_val == self._is_online():
            self._warn('Want to set online state to current state!')
            return True
        if new_val:
            self._log.info('going online')
            args = [self._tell_connd_bin, '--online', 'P1']
        else:
            self._log.info('going offline')
            args = [self._tell_connd_bin, '--offline']

        try:
            return_code = subprocess.call(args, stderr=subprocess.STDOUT, timeout=1)
            if return_code == 0:
                self._prev_state_changes.append((new_val, dt.now()))
                return True
            else:
                self._log.info('tell-connd returned {0} --> online state probably unchanged!'.format(\
                                return_code))
                return False
        except subprocess.TimeoutExpired as te:
            self._warn('Timeout expired running {0}. Output was {1}'.format(args, te.output))
            return False
        except:
            self._warn('Trouble calling tell-connd', exc_info=sys.exc_info())
            return False
    #end: Tester._set_online


    def _ensure_all_pinged(self):
        """ called before going offline; check if all hosts were successfully pinged """
        # determine when we last went online
        time_went_online = None
        time_go_offline = dt.now()

        for went_online, change_time in reversed(self._prev_state_changes):
            if went_online:
                time_went_online = change_time
                break
        if time_went_online is None:
            # never been online --> nothing to check yet
            return

        self._log.info('checking whether all hosts were pinged between '
                     + '{0:%H:%M:%S} and now ({1:%H:%M:%S})'.format(\
                        time_went_online, time_go_offline))
        for ip, reply_time in self._last_reply.items():
            if reply_time < time_went_online:
                self._log.warning(\
                        'Did not ping host {0} between {1:%H:%M:%S} and now ({2:%H:%M:%S})!'.format(\
                        ip, time_went_online, time_go_offline))
                self._fails.append((ip, time_went_online, time_go_offline))
                self._log.info('remembered as failure number {0} / {1}'.format(\
                        len(self._fails), self.max_fails))

                # check if there were too many fails; if so, stop
                if len(self._fails) > self.max_fails:
                    self._fails.append(("WARN", dt.now(), "Too many failurs/warnings!"))
                    raise TooManyWarnings()
            #end: if not pinged recently
        #end: for all known ips
    #end: Tester._ensure_all_pinged


    def _warn(self, message, **kwargs):

        self._log.warning(message, **kwargs)
        self._fails.append(("WARN", dt.now(), message))
        self._log.info('remembered as failure number {0} / {1}'.format(\
                len(self._fails), self.max_fails))

        # check if there were too many fails; if so, stop
        if len(self._fails) > self.max_fails:
            self._fails.append(("WARN", dt.now(), "Too many failurs/warnings!"))
            raise TooManyWarnings()
    #end: Tester._warn


    #def _print(self, message):
    #    now = dt.now()
    #    if self._next_state_change is not None:
    #        print('{0:%H:%M:%S};next{1:+5.1f}s;{2} fails: {3}'.format(now, \
    #                (now - self._next_state_change).total_seconds(), len(self._fails), message))
    #    else:
    #        print('{0:%H:%M:%S};next-?????;{1} fails: {2}'.format(now, len(self._fails), message))


    def print_failures(self, log):
        if len(self._fails) == 0:
            log.info('No fails! :-)')
        else:
            log.info('Recorded the following failures and warnings:')
            for ip, time_went_online, time_go_offline in self._fails:
                if ip == 'WARN':
                    log.info('Warning at {0:%H:%M:%S}: {1}'.format(time_went_online, time_go_offline))
                else:
                    log.info('Failed to ping {0} between {1:%H:%M:%S} and {2:%H:%M:%S} ({3:.2f}s)'\
                            .format(ip, time_went_online, time_go_offline, \
                                    (time_go_offline-time_went_online).total_seconds()))
#end: class Tester

########################################################################
### MAIN
########################################################################

def parse_args():
    """ parse command line arguments and set defaults """
    from argparse import ArgumentParser
    parser = ArgumentParser(description="Long-term testing of pingcheck: goes online and offline " \
                                      + "using tell-connd and checks log for correct output", \
                            epilog="(c) Christian Herdtweck, Intra2net AG, 2015")

    # binaries
    parser.add_argument('--pingcheck-binary', '-p', metavar='PATH', \
            default=default_pingcheck_bin, \
            help='path to pingcheck binary')
    parser.add_argument('--tell-connd-binary', '-t', metavar='PATH', \
            default=default_tell_connd_bin, \
            help='path to tell-connd binary')

    # output
    parser.add_argument('--output-prefix', '-o', metavar='PATH', \
            default=default_output_prefix, \
            help='path and/or file prefix to config and log files; ~=home, 0=prog start datetime; ' \
                + 'default="%(default)s"')
    parser.add_argument('--log-to-console', '-c', metavar='0|1', \
            choices=(0, 1), type=int, default=1, \
            help='print all output to console? default=%(default)s')
    parser.add_argument('--log-level-console', metavar='LEVEL', \
            type=int, default=logging.INFO, \
            help='log level for console output; ' \
               + 'debug={0}, info={1}, warn={2}; default: %(default)s'.format(\
                logging.DEBUG, logging.INFO, logging.WARN))
    parser.add_argument('--log-to-file', '-f', metavar='0|1', \
            choices=(0, 1), type=int, default=1, \
            help='print all output to log file? default=%(default)s')
    parser.add_argument('--log-level-file', metavar='LEVEL', \
            type=int, default=logging.DEBUG, \
            help='log level for console output; ' \
               + 'debug={0}, info={1}, warn={2}; default: %(default)s'.format(\
                logging.DEBUG, logging.INFO, logging.WARN))
    parser.add_argument('--log-file-suffix', '-l', metavar='SUFF.IX', \
            type=str, default='.log', \
            help='2nd part of log file name after --output_prefix; default="%(default)s"')
    parser.add_argument('--pinger-log-level', metavar='LEVEL', \
            type=str, default='INFO', \
            help='log level given to pingchecker, DEBUG might quickly fill log; default=%(default)s')


    parser.add_argument('--max-fails-per-host', '-m', metavar='N', type=int, default=10, \
            help='max number of failures/warnings per host before test will end; default: %(default)s')
    parser.add_argument('--version', action='version', version='%(prog)s {0}'.format(version))

    # parse
    opts = parser.parse_args()
    opts.output_prefix = expanduser(opts.output_prefix).format(dt.now())
    return opts
#end: parse_args


def setup_logging(opts):
    """ create main logger and handlers for base logger """

    # create logger for main
    log = logger_base.getChild('main')

    # remove all existing handlers
    for old_handler in logger_base.handlers:
        print('removing handler from base')
        logger_base.removeHandler(old_handler)
    for old_handler in log.handlers:
        print('removing handler from log')
        log.removeHandler(old_handler)

    logger_base.setLevel(logging.DEBUG)

    # create handlers and format for base logger
    log_file = None
    if opts.log_to_console:
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(opts.log_level_console)
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s %(message)s', \
                                      datefmt='%d.%m,%H:%M:%S')
        handler.setFormatter(formatter)
        logger_base.addHandler(handler)

    if opts.log_to_file:
        log_file = opts.output_prefix + opts.log_file_suffix
        handler = logging.FileHandler(log_file)
        handler.setLevel(opts.log_level_file)
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s', \
                                      datefmt='%d.%m,%H:%M:%S')
        handler.setFormatter(formatter)
        logger_base.addHandler(handler)

    # log info
    log.info('Log for {0} in {1} created on {2}'.format(__name__, __file__, dt.now()))
    if opts.log_to_console:
        log.info('logging to console at level {0}'.format(opts.log_level_console))
    else:
        log.info('not logging to console')
    if opts.log_to_file:
        log.info('logging to file {1} at level {0}'.format(opts.log_level_file, log_file))
    else:
        log.info('not logging to file')

    # return log_file only if makes sense to talk about it in main
    if opts.log_to_console:
        return log, log_file
    else:
        return log, None
#end: setup_logging


def handle_sigchild(log, pinger, tester, signum, curr_frame):
    print("\n!received sigchild!\n")
    log.error('Received SIGCHLD')
    log.info('with args {0} and {1}'.format(signum, curr_frame))

    cleanup(log, pinger, tester)


def cleanup(log, pinger, tester, line_iter):
    """ the stuff from finally clause in main loop; also called from handle_sigchild """

    return_code = 255

    log.info('-' * 72)
    if pinger is None:
        log.warning('Error before pinger was created')
        return_code = 254
    else:
        log.info("Cleaning up")

        for line in line_iter:
            log.info('left-over output: {0}'.format(line))

        try:
            log.info('telling pinger to terminate')
            pinger.terminate()

            #return_code = pinger.wait(3)  may block!
            outs, errs = pinger.communicate(timeout=3)
            for line in outs.splitlines():
                log.info('left-over output: {0}'.format(line))
            for line in errs.splitlines():
                log.info('left-over output: {0}'.format(line))

            return_code = pinger.returncode   # may still be None
            log.info('pinger terminated with return code {0}'.format(return_code))
        except subprocess.TimeoutExpired as te:
            log.warning('Timeout expired waiting for pinger to terminate: {0}'.format(te))
            return_code = 253
        except:
            log.exception('trouble terminating pinger', exc_info=sys.exc_info())

        if return_code is None:
            for line in line_iter:
                log.info('left-over output: {0}'.format(line))

            # kill it
            try:
                log.info('killing pinger the hard way')
                pinger.kill()
                outs, errs = pinger.communicate(timeout=3)
                for line in outs.splitlines():
                    log.info('left-over output: {0}'.format(line))
                for line in errs.splitlines():
                    log.info('left-over output: {0}'.format(line))

            except:
                log.exception('trouble killing pinger', exc_info=sys.exc_info())

        if tester is not None:
            tester.print_failures(log)
    #end: if have pinger

    # reset original signal handler
    #signal.signal(signal.SIGCHLD, signal.SIG_DFL)

    return return_code
#end: cleanup


def main():
    """ parses args, create a Pinger and Parser and Tester and runs them; cleans up
    """

    opts = parse_args()

    base_dir = dirname(opts.output_prefix + "_test")
    if not isdir(base_dir):
        mkdir(base_dir)

    log, log_file = setup_logging(opts)

    pinger = None
    return_code = 255
    tester = None
    line_iter = None
    try:
        config_file, n_hosts = create_conf_file(opts.output_prefix)
        log.info('created config file {0}'.format(config_file))

        #signal.signal(signal.SIGCHLD, \
        #        lambda signum, curr_frame: handle_sigchild(log, pinger, tester, signum, curr_frame))

        # pingcheck subprocess
        pinger = subprocess.Popen([opts.pingcheck_binary, '--log-level', opts.pinger_log_level, \
                                     '--log-output', 'TERMINAL', '--config-file', config_file], \
                                   stdout=subprocess.PIPE, stderr=subprocess.PIPE, \
                                   universal_newlines=True)
        log.info('started pingcheck with PID {0}'.format(pinger.pid))

        parser = ProcOutputParser(pinger.stdout, pinger.stderr)
        line_iter = iter(parser)
        tester = Tester(line_iter, opts.tell_connd_binary, max(10, opts.max_fails_per_host*n_hosts))

        tester.test()
    except KeyboardInterrupt:
        log.info('stopped by user')
    except:
        log.exception('Caught an exception in main loop', exc_info=sys.exc_info())
        log.warning('Re-raise')
        raise
    finally:
        return_code = cleanup(log, pinger, tester, line_iter)

        log.info('Will stop now')
        if log_file is not None:
            log.info('might find more log output in {0}'.format(log_file))
        log.info('-' * 72)
    #end: global clean-up

    return return_code

#end: function main


if __name__ == '__main__':
    main()
    logging.shutdown()

# (created using vim -- the world's best text editor)

